#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import inspect

import numpy as np
from grakn.client import GraknClient

from kglib.utils.grakn.synthetic.statistics.pmf import PMF


def get_example_queries(pmf, example_id):

    variable_values = pmf.select()

    queries = [f'insert $p isa person, has example-id {example_id};',
               f'insert $doc isa person, has example-id {20000 + example_id};']

    if variable_values['Multiple Sclerosis'] is not False:
        queries.append(inspect.cleandoc(f'''match
                       $d isa disease, has name "Multiple Sclerosis";
                       $p isa person, has example-id {example_id};
                       $doc isa person, has example-id {20000 + example_id};
                       insert
                       (patient: $p, diagnosed-disease: $d, doctor: $doc) isa diagnosis;
                       $p has age {int(variable_values['Multiple Sclerosis']['age']())};'''))

    if variable_values['Diabetes Type II'] is not False:
        queries.append(inspect.cleandoc(f'''match
                       $p isa person, has example-id {example_id};
                       $d isa disease, has name "Diabetes Type II";
                       $doc isa person, has example-id {20000 + example_id};
                       insert
                       (patient: $p, diagnosed-disease: $d, doctor: $doc) isa diagnosis;
                       $p has age {int(variable_values['Diabetes Type II']['age']())};'''))

    if variable_values['Fatigue'] is not False:
        queries.append(inspect.cleandoc(f'''match
                       $p isa person, has example-id {example_id};
                       $s isa symptom, has name "Fatigue";
                       insert
                       (presented-symptom: $s, symptomatic-patient: $p) isa 
                       symptom-presentation, has severity {variable_values['Fatigue']['severity']()};'''))

    if variable_values['Blurred vision'] is not False:
        queries.append(inspect.cleandoc(f'''match
                       $p isa person, has example-id {example_id};
                       $s isa symptom, has name "Blurred vision";
                       insert
                       (presented-symptom: $s, symptomatic-patient: $p) isa 
                       symptom-presentation, has severity {variable_values['Blurred vision']['severity']()};'''))

    if variable_values['Drinking'] is not False:
        queries.append(inspect.cleandoc(f'''match
                       $p isa person, has example-id {example_id};
                       $s isa substance, has name "Alcohol";
                       insert
                       $c(consumer: $p, consumed-substance: $s) isa consumption, 
                       has units-per-week {int(variable_values['Drinking']['units-per-week']())};'''))

    if variable_values['Parent has Diabetes Type II'] is not False:
        queries.append(inspect.cleandoc(f'''match
                       $p isa person, has example-id {example_id};
                       $d isa disease, has name "Diabetes Type II";
                       insert
                       (parent: $parent, child: $p) isa parentship;
                       $parent isa person, has example-id {example_id + 10000};
                       (patient: $parent, diagnosed-disease: $d) isa diagnosis;
                       '''))

    if variable_values['Cigarettes'] is not False:
        queries.append(inspect.cleandoc(f'''match
                       $p isa person, has example-id {example_id};
                       $s isa substance, has name "Cigarettes";
                       insert
                       $c(consumer: $p, consumed-substance: $s) isa consumption, 
                       has units-per-week {int(variable_values['Cigarettes']['units-per-week']())};'''))

    return queries


def generate_example_graphs(num_examples, keyspace="diagnosis", uri="localhost:48555"):

    client = GraknClient(uri=uri)
    session = client.session(keyspace=keyspace)

    pmf_array = np.zeros([2, 2, 2, 2, 3, 2, 3], dtype=np.float)
    pmf_array[0, 1, 0, 1, 0, 0, 0] = 0.1
    pmf_array[1, 0, 1, 0, 0, 0, 0] = 0.05
    pmf_array[1, 0, 1, 0, 2, 0, 0] = 0.1
    pmf_array[0, 1, 1, 0, 0, 0, 0] = 0.05
    pmf_array[1, 0, 0, 1, 0, 0, 0] = 0.19
    pmf_array[1, 0, 0, 1, 0, 1, 0] = 0.15
    pmf_array[1, 1, 1, 1, 0, 0, 0] = 0.01
    pmf_array[0, 1, 1, 1, 0, 0, 0] = 0.05
    pmf_array[0, 1, 1, 1, 0, 0, 1] = 0.05
    pmf_array[0, 1, 1, 1, 0, 0, 2] = 0.1
    pmf_array[1, 0, 1, 1, 0, 0, 0] = 0.05
    pmf_array[1, 0, 1, 1, 2, 1, 2] = 0.1

    def normal_dist(mean, var):
        return lambda: round(np.random.normal(mean, var, 1)[0], 2)

    pmf = PMF({
        'Diabetes Type II':             [False, {'age': normal_dist(60, 10)}],
        'Multiple Sclerosis':           [False, {'age': normal_dist(30, 10)}],
        'Fatigue':                      [False, {'severity': normal_dist(0.3, 0.1)}],
        'Blurred vision':               [False, {'severity': normal_dist(0.5, 0.2)}],
        'Drinking':                     [False, {'units-per-week': normal_dist(5, 1)}, {'units-per-week': normal_dist(20, 3)}],
        'Parent has Diabetes Type II':  [False, True],
        'Cigarettes':                   [False, {'units-per-week': normal_dist(5, 1)}, {'units-per-week': normal_dist(20, 3)}],
    }, pmf_array, seed=0)

    # print(pmf.to_dataframe()) # TODO Remove pandas if this is not needed now

    for example_id in range(0, num_examples):
        tx = session.transaction().write()
        for query in get_example_queries(pmf, example_id):
            print(query)
            tx.query(query)
        tx.commit()

    session.close()
    client.close()


if __name__ == '__main__':
    generate_example_graphs(100, keyspace="diagnosis", uri="localhost:48555")



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import numpy as np

from kglib.utils.graph.iterate import multidigraph_data_iterator, multidigraph_node_data_iterator, \
    multidigraph_edge_data_iterator


def encode_values(graph, categorical_attributes, continuous_attributes):
    for node_data in multidigraph_node_data_iterator(graph):
        typ = node_data['type']

        if categorical_attributes is not None and typ in categorical_attributes.keys():
            # Add the integer value of the category for each categorical attribute instance
            category_values = categorical_attributes[typ]
            node_data['encoded_value'] = category_values.index(node_data['value'])

        elif continuous_attributes is not None and typ in continuous_attributes.keys():
            min_val, max_val = continuous_attributes[typ]
            node_data['encoded_value'] = (node_data['value'] - min_val) / (max_val - min_val)

        else:
            node_data['encoded_value'] = 0
    for edge_data in multidigraph_edge_data_iterator(graph):
        edge_data['encoded_value'] = 0

    return graph


def encode_types(graph, iterator_func, types):
    """
    Encodes the type found in graph data as an integer according to the index it is found in `all_types`
    Args:
        graph: The graph to encode
        iterator_func: An function to create an iterator of data in the graph (node data, edge data or combined node and edge data)
        types: The full list of types to be encoded in this order
        
    Returns:
        The graph, which is also is updated in-place

    """
    iterator = iterator_func(graph)

    for data in iterator:
        data['categorical_type'] = types.index(data['type'])

    return graph


def create_input_graph(graph):
    input_graph = graph.copy()

    for data in multidigraph_data_iterator(input_graph):
        if data["solution"] == 0:
            preexists = 1
        else:
            preexists = 0

        features = stack_features([preexists, data["categorical_type"], data["encoded_value"]])
        data.clear()
        data["features"] = features

    input_graph.graph["features"] = np.array([0.0] * 5, dtype=np.float32)
    return input_graph


def create_target_graph(graph):
    target_graph = graph.copy()
    solution_one_hot_encoding = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], dtype=np.float32)

    for data in multidigraph_data_iterator(target_graph):
        features = solution_one_hot_encoding[data["solution"]]
        data.clear()
        data["features"] = features

    target_graph.graph["features"] = np.array([0.0] * 5, dtype=np.float32)
    return target_graph


def stack_features(features):
    """
    Stacks features together into a single vector

    Args:
        features: iterable of features, features can be a single value or iterable

    Returns:
        Numpy array (vector) of stacked features

    """

    return np.hstack([np.array(feature, dtype=np.float32) for feature in features])



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#


def get_call_args(mock):
    """
    Get the arguments used to call a mock for each call to the mock. Necessary since `.assert_has_calls` won't work
    for numpy arrays
    Args:
        mock: the mock

    Returns:
        A list of lists. The outer list is the calls made, the inner list is the arguments given for that call
    """
    flat_args = []
    args_list = mock.call_args_list
    for call in args_list:
        args, kwargs = call
        flat_args.append(args)
    return flat_args



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from setuptools import setup
from setuptools import find_packages

packages = find_packages()

setup(
    name = "grakn-kglib",
    version = "0.2.1",
    description = "A Machine Learning Library for the Grakn knowledge graph.",
    long_description = open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers = ["Programming Language :: Python :: 3", "Programming Language :: Python :: 3.6", "License :: OSI Approved :: Apache Software License", "Operating System :: OS Independent", "Intended Audience :: Developers", "Intended Audience :: Science/Research", "Topic :: Scientific/Engineering", "Topic :: Scientific/Engineering :: Information Analysis", "Topic :: Scientific/Engineering :: Artificial Intelligence", "Topic :: Software Development :: Libraries", "Topic :: Software Development :: Libraries :: Python Modules"],
    keywords = "machine learning logical reasoning knowledege graph grakn database graph knowledgebase knowledge-engineering",
    url = "https://github.com/graknlabs/kglib",
    author = "Grakn Labs",
    author_email = "community@grakn.ai",
    license = "Apache-2.0",
    packages=packages,
    install_requires=["enum-compat==0.0.2", "grakn-client==1.6.0", "absl-py==0.8.0", "astor==0.8.0", "cloudpickle==1.2.2", "contextlib2==0.5.5", "cycler==0.10.0", "decorator==4.4.0", "dm-sonnet==1.35", "future==0.17.1", "gast==0.3.1", "google-pasta==0.1.7", "graph-nets==1.0.4", "grpcio==1.23.0", "h5py==2.10.0", "Keras-Applications==1.0.8", "Keras-Preprocessing==1.1.0", "kiwisolver==1.1.0", "Markdown==3.1.1", "matplotlib==3.1.1", "networkx==2.3", "numpy==1.17.2", "pandas==0.25.1", "protobuf==3.9.1", "pyparsing==2.4.2", "python-dateutil==2.8.0", "pytz==2019.2", "scipy==1.3.1", "semantic-version==2.8.2", "six==1.12.0", "tensorboard==1.14.0", "tensorflow==1.14.0", "tensorflow-estimator==1.14.0", "tensorflow-probability==0.7.0", "termcolor==1.1.0", "Werkzeug==0.15.6", "wrapt==1.11.2"],
    zip_safe=False,
)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import datetime
import six
from grakn.service.Session.util import enums
from grakn.service.Session.Concept import ConceptFactory
from grakn.exception.GraknError import GraknError 


class ResponseReader(object):
    
    @staticmethod
    def query(tx_service, grpc_query_iter):
        iterator_id = grpc_query_iter.id
        return ResponseIterator(tx_service,
                                iterator_id,
                                lambda tx_serv, iterate_res: AnswerConverter.convert(tx_serv, iterate_res.query_iter_res.answer))


    @staticmethod
    def get_concept(tx_service, grpc_get_schema_concept):
        which_one = grpc_get_schema_concept.WhichOneof("res")
        if which_one == "concept":
            grpc_concept = grpc_get_schema_concept.concept
            return ConceptFactory.create_concept(tx_service, grpc_concept)
        elif which_one == "null":
            return None
        else:
            raise GraknError("Unknown get_concept response: {0}".format(which_one))

    @staticmethod
    def get_schema_concept(tx_service, grpc_get_concept):
        which_one = grpc_get_concept.WhichOneof("res")
        if which_one == "schemaConcept":
            grpc_concept = grpc_get_concept.schemaConcept
            return ConceptFactory.create_concept(tx_service, grpc_concept)
        elif which_one == "null":
            return None
        else:
            raise GraknError("Unknown get_schema_concept response: {0}".format(which_one))

    @staticmethod
    def get_attributes_by_value(tx_service, grpc_get_attrs_iter):
        iterator_id = grpc_get_attrs_iter.id
        return ResponseIterator(tx_service,
                                iterator_id,
                                lambda tx_serv, iterate_res: ConceptFactory.create_concept(tx_serv, iterate_res.getAttributes_iter_res.attribute))

    @staticmethod
    def put_entity_type(tx_service, grpc_put_entity_type):
        return ConceptFactory.create_concept(tx_service, grpc_put_entity_type.entityType) 

    @staticmethod
    def put_relation_type(tx_service, grpc_put_relation_type):
        return ConceptFactory.create_concept(tx_service, grpc_put_relation_type.relationType)

    @staticmethod
    def put_attribute_type(tx_service, grpc_put_attribute_type):
        return ConceptFactory.create_concept(tx_service, grpc_put_attribute_type.attributeType)

    @staticmethod
    def put_role(tx_service, grpc_put_role):
        return ConceptFactory.create_concept(tx_service, grpc_put_role.role)

    @staticmethod
    def put_rule(tx_service, grpc_put_rule):
        return ConceptFactory.create_concept(tx_service, grpc_put_rule.rule)

    @staticmethod
    def from_grpc_value_object(grpc_value_object):
        whichone = grpc_value_object.WhichOneof('value')
        # check the one is in the known datatypes
        known_datatypes = [e.name.lower() for e in enums.DataType]
        if whichone.lower() not in known_datatypes:
            raise GraknError("Unknown value object value key: {0}, not in {1}".format(whichone, known_datatypes))
        if whichone == 'string':
            return grpc_value_object.string
        elif whichone == 'boolean':
            return grpc_value_object.boolean
        elif whichone == 'integer':
            return grpc_value_object.integer
        elif whichone == 'long':
            return grpc_value_object.long
        elif whichone == 'float':
            return grpc_value_object.float
        elif whichone == 'double':
            return grpc_value_object.double
        elif whichone == 'date':
            epoch_ms_utc = grpc_value_object.date
            local_datetime_utc = datetime.datetime.fromtimestamp(float(epoch_ms_utc)/1000.)
            return local_datetime_utc
        else:
            raise GraknError("Unknown datatype in enum but not handled in from_grpc_value_object")
        

    # --- concept method helpers ---

    @staticmethod
    def iter_res_to_iterator(tx_service, iterator_id, next_iteration_handler):
        return ResponseIterator(tx_service, iterator_id, next_iteration_handler)

    @staticmethod
    def create_explanation(tx_service, grpc_explanation_res):
        """ Convert gRPC explanation response to explanation object """
        grpc_list_of_concept_maps = grpc_explanation_res.explanation
        native_list_of_concept_maps = []
        for grpc_concept_map in grpc_list_of_concept_maps:
            native_list_of_concept_maps.append(AnswerConverter._create_concept_map(tx_service, grpc_concept_map))
        return Explanation(native_list_of_concept_maps)

class Explanation(object):

    def __init__(self, list_of_concept_maps):
        self._concept_maps_list = list_of_concept_maps

    def get_answers(self):
        """ Return answers this explanation is dependent on"""
        # note that concept_maps are subtypes of Answer
        return self._concept_maps_list


# ----- Different types of answers -----

class AnswerGroup(object):

    def __init__(self, owner_concept, answer_list):
        self._owner_concept = owner_concept
        self._answer_list = answer_list

    def get(self):
        return self

    def owner(self):
        return self._owner_concept

    def answers(self):
        return self._answer_list


class ConceptMap(object):

    def __init__(self, concept_map, query_pattern, has_explanation, tx_service):
        self._concept_map = concept_map
        self._has_explanation = has_explanation
        self._query_pattern = query_pattern
        self._tx_service = tx_service

    def get(self, var=None):
        """ Get the indicated variable's Concept from the map or this ConceptMap """
        if var is None:
            return self
        else:
            if var not in self._concept_map:
                # TODO specialize exception
                raise GraknError("Variable {0} is not in the ConceptMap".format(var))
            return self._concept_map[var]

    def query_pattern(self):
        return self._query_pattern

    def has_explanation(self):
        return self._has_explanation

    def explanation(self):
        if self._has_explanation:
            return self._tx_service.explanation(self)
        else:
            raise GraknError("Explanation not available on concept map: " + str(self))

    def map(self):
        """ Get the map from Variable (str) to Concept objects """
        return self._concept_map

    def vars(self):
        """ Get a set of vars in the map """
        return set(self._concept_map.keys())

    def contains_var(self, var):
        """ Check whether the map contains the var """
        return var in self._concept_map

    def is_empty(self):
        """ Check if the variable map is empty """
        return len(self._concept_map) == 0


class ConceptList(object):

    def __init__(self, concept_id_list):
        self._concept_id_list = concept_id_list

    def list(self):
        """ Get the list of concept IDs """
        return self._concept_id_list


class ConceptSet(object):

    def __init__(self, concept_id_set):
        self._concept_id_set = concept_id_set
    __init__.__annotations__ = {'_concept_id_set': 'List[str]'}

    def set(self):
        """ Return the set of Concept IDs within this ConceptSet """
        return self._concept_id_set


class ConceptSetMeasure(ConceptSet):

    def __init__(self, concept_id_set, number):
        super(ConceptSetMeasure, self).__init__(concept_id_set)
        self._measurement = number
    __init__.__annotations__ = {'_measurement': float}

    def measurement(self):
        return self._measurement


class Value(object):

    def __init__(self, number):
        self._number = number
    __init__.__annotations__ = {'number': float}

    def number(self):
        """ Get as number (float or int) """
        return self._number


class Void(object):
    def __init__(self, message):
        self._message = message
    __init__.__annotations__ = {'message': str}

    def message(self):
        """ Get the message on this Void answer type """
        return self._message


class AnswerConverter(object):
    """ Static methods to convert answers into Answer objects """

    @staticmethod
    def convert(tx_service, grpc_answer):
        which_one = grpc_answer.WhichOneof('answer')

        if which_one == 'conceptMap':
            return AnswerConverter._create_concept_map(tx_service, grpc_answer.conceptMap)
        elif which_one == 'answerGroup':
            return AnswerConverter._create_answer_group(tx_service, grpc_answer.answerGroup)
        elif which_one == 'conceptList':
            return AnswerConverter._create_concept_list(tx_service, grpc_answer.conceptList)
        elif which_one == 'conceptSet':
            return AnswerConverter._create_concept_set(tx_service, grpc_answer.conceptSet)
        elif which_one == 'conceptSetMeasure':
            return AnswerConverter._create_concept_set_measure(tx_service, grpc_answer.conceptSetMeasure)
        elif which_one == 'value':
            return AnswerConverter._create_value(tx_service, grpc_answer.value)
        elif which_one == 'void':
            return AnswerConverter._create_void(tx_service, grpc_answer.void)
        else:
            raise GraknError('Unknown gRPC Answer.answer message type: {0}'.format(which_one))
   
    @staticmethod
    def _create_concept_map(tx_service, grpc_concept_map_msg):
        """ Create a Concept Dictionary from the grpc response """
        var_concept_map = grpc_concept_map_msg.map
        answer_map = {}
        for (variable, grpc_concept) in var_concept_map.items():
            answer_map[variable] = ConceptFactory.create_concept(tx_service, grpc_concept)

        query_pattern = grpc_concept_map_msg.pattern
        has_explanation = grpc_concept_map_msg.hasExplanation

        return ConceptMap(answer_map,  query_pattern, has_explanation, tx_service)

    @staticmethod
    def _create_answer_group(tx_service, grpc_answer_group):
        grpc_owner_concept = grpc_answer_group.owner
        owner_concept = ConceptFactory.create_concept(tx_service, grpc_owner_concept)
        grpc_answers = list(grpc_answer_group.answers)
        answer_list = [AnswerConverter.convert(tx_service, grpc_answer) for grpc_answer in grpc_answers]
        return AnswerGroup(owner_concept, answer_list)

    @staticmethod
    def _create_concept_list(tx_service, grpc_concept_list_msg):
        ids_list = list(grpc_concept_list_msg.list.ids)
        return ConceptList(ids_list)

    @staticmethod
    def _create_concept_set(tx_service, grpc_concept_set_msg):
        ids_set = set(grpc_concept_set_msg.set.ids)
        return ConceptSet(ids_set)

    @staticmethod
    def _create_concept_set_measure(tx_service, grpc_concept_set_measure):
        concept_ids = list(grpc_concept_set_measure.set.ids)
        number = grpc_concept_set_measure.measurement.value 
        return ConceptSetMeasure(concept_ids, AnswerConverter._number_string_to_native(number))

    @staticmethod
    def _create_value(tx_service, grpc_value_msg):
        number = grpc_value_msg.number.value 
        return Value(AnswerConverter._number_string_to_native(number))

    @staticmethod
    def _create_void(tx_service, grpc_void):
        """ Convert grpc Void message into an object """
        return Void(grpc_void.message)

    @staticmethod
    def _number_string_to_native(number):
        try:
            return int(number)
        except ValueError:
            return float(number)


class ResponseIterator(six.Iterator):
    """ Retrieves next value in the Grakn response iterator """

    def __init__(self, tx_service , iterator_id, iter_resp_converter):
        self._tx_service = tx_service  
        self.iterator_id = iterator_id
        self._iter_resp_converter = iter_resp_converter

    def __iter__(self):
        return self

    def __next__(self):
        # get next from server
        iter_response = self._tx_service.iterate(self.iterator_id)
        # print("Iterator response:")
        # print(iter_response)
        which_one = iter_response.WhichOneof("res")
        if which_one == 'done' and iter_response.done:
            raise StopIteration()
        else:
            return self._iter_resp_converter(self._tx_service, iter_response)

    def collect_concepts(self):
        """ Helper method to retrieve concepts from a query() method """
        concepts = []
        for answer in self:
            if not isinstance(answer, ConceptMap):
                raise GraknError("Only use .collect_concepts on ConceptMaps returned by query()")
            concepts.extend(answer.map().values()) # get concept map => concepts
        return concepts








#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from grakn.service.Session.Concept import BaseTypeMapping
from grakn.service.Session.Concept.Concept import EntityType, RelationType, AttributeType, Role, Rule, Entity, Relation, Attribute, Type


# map names to ConceptHierarchy types
name_to_object = {
    BaseTypeMapping.META_TYPE: Type,
    BaseTypeMapping.ENTITY_TYPE: EntityType,
    BaseTypeMapping.RELATION_TYPE: RelationType,
    BaseTypeMapping.ATTRIBUTE_TYPE: AttributeType,
    BaseTypeMapping.ROLE: Role,
    BaseTypeMapping.RULE: Rule,
    BaseTypeMapping.ENTITY: Entity,
    BaseTypeMapping.RELATION: Relation,
    BaseTypeMapping.ATTRIBUTE: Attribute
}


def create_concept(tx_service, grpc_concept):
    """ Instantate a local Python object for a Concept corresponding to the .baseType of the GRPC concept object """

    concept_id = grpc_concept.id
    base_type = grpc_concept.baseType

    try:
        concept_name = BaseTypeMapping.grpc_base_type_to_name[base_type]
        concept_class = name_to_object[concept_name]
    except KeyError as ke:
        raise ke

    return concept_class(concept_id, concept_name, tx_service)



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import numpy as np
import pandas as pd


class PMF:
    def __init__(self, variables, pmf_array, seed=None):
        """
        Probability Mass Function, the discrete equivalent of a Joint Probability Density Function

        Args:
            variables: An ordered dict, keys: the named variables of the function; values: the possible (discrete)
                values for each variable
            pmf_array: Probability Mass Function array: A numpy nd array of probabilities that corresponds with
                `variables`. Each variable is represented by a dimension of the array. For example, when there are 3
                variables, element (0, 0, 0) indicates the probability that all variables take the first value given
                for them in `variables`

        Raises:
            IndexError if `variables` and `pmf_array` are inconsistent
        """
        self._variables = variables
        self._pmf_array = pmf_array

        # Check that `self._variables` and `self._pmf_array` are consistent
        values_shape = tuple(
            len(discrete_values) for i, (variable, discrete_values) in enumerate(self._variables.items()))

        if values_shape != self._pmf_array.shape:
            raise IndexError(f'Variable values have combined shape {values_shape}, whereas the PMF array given has '
                             f'shape {self._pmf_array.shape}')

        if seed is not None:
            np.random.seed(seed)

    def select(self):
        """
        Select a set of variable values from the PMF, using the probabilities supplied in `pmf_array` as weights.

        Returns:
            A dict key: variable names; values: the chosen value of each variable
        """
        flattened = self._pmf_array.flatten()

        answer = {}

        indices = list(np.ndindex(self._pmf_array.shape))
        int_index = list(range(len(indices)))
        chosen_int = np.random.choice(int_index, p=flattened)
        chosen_index = indices[chosen_int]
        for index, (variable, discrete_values) in zip(chosen_index, self._variables.items()):
            answer[variable] = discrete_values[index]
        return answer

    def to_dataframe(self):
        """
        Creates a DataFrame of the PMF, most useful for visualisation purposes

        Returns:
            A pandas DataFrame, multi-indexed by the variables and their possible values

        """
        variables = list(self._variables.keys())
        variable_values = list(self._variables.values())
        index = pd.MultiIndex.from_product(variable_values, names=variables)

        return pd.DataFrame(self._pmf_array.flatten(), index=index)



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import numpy as np
import tensorflow as tf


def loss_ops_from_difference(target_op, output_ops):
    """
    Loss operation which directly compares the target with the output over all nodes and edges
    Args:
        target_op: The target of the model
        output_ops: A list of the outputs of the model, one for each message-passing step

    Returns: The loss for each message-passing step

    """
    loss_ops = [
        tf.losses.softmax_cross_entropy(target_op.nodes, output_op.nodes)
        for output_op in output_ops
    ]
    return loss_ops


def loss_ops_preexisting_no_penalty(target_op, output_ops):
    """
    Loss operation which doesn't penalise the output values for pre-existing nodes and edges, treating them as slack
    variables

    Args:
        target_op: The target of the model
        output_ops: A list of the outputs of the model, one for each message-passing step

    Returns: The loss for each message-passing step

    """
    loss_ops = []
    for output_op in output_ops:
        node_mask_op = tf.math.reduce_any(
            tf.math.not_equal(target_op.nodes, tf.constant(np.array([1., 0., 0.]), dtype=tf.float32)), axis=1)
        target_nodes = tf.boolean_mask(target_op.nodes, node_mask_op)
        output_nodes = tf.boolean_mask(output_op.nodes, node_mask_op)

        loss_op = tf.losses.softmax_cross_entropy(target_nodes, output_nodes)

        loss_ops.append(loss_op)

    return loss_ops


#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#


class MockConcept:
    def __init__(self, id):
        self.id = id


class MockType(MockConcept):
    def __init__(self, id, label, base_type):
        super().__init__(id)
        self._label = label
        assert base_type in {'ENTITY', 'RELATION', 'ATTRIBUTE'}
        self.base_type = base_type

    def label(self):
        return self._label


class DataType:
    def __init__(self, name):
        self.name = name


class MockAttributeType(MockType):
    def __init__(self, id, label, base_type, data_type):
        super().__init__(id, label, base_type)
        assert data_type in {'STRING', 'LONG', 'DOUBLE', 'DATE', 'BOOLEAN'}
        self._data_type = DataType(data_type)

    def data_type(self):
        return self._data_type


class MockThing(MockConcept):
    def __init__(self, id, type):
        super().__init__(id)
        assert(isinstance(type, MockType))
        self._type = type

    def type(self):
        return self._type

    @property
    def base_type(self):
        return self._type.base_type


class MockAttribute(MockThing):
    def __init__(self, id, value, type):
        super().__init__(id, type)
        self._value = value
        assert isinstance(type, MockAttributeType)

    def value(self):
        return self._value



#
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#

import tensorflow as tf
import sonnet as snt

from kglib.kgcn.models.attribute import CategoricalAttribute, ContinuousAttribute, BlankAttribute
from kglib.kgcn.models.typewise import TypewiseEncoder


class ThingEmbedder(snt.AbstractModule):
    def __init__(self, node_types, type_embedding_dim, attr_embedding_dim, categorical_attributes,
                 continuous_attributes, name="ThingEmbedder"):
        super(ThingEmbedder, self).__init__(name=name)

        self._node_types = node_types
        self._type_embedding_dim = type_embedding_dim
        self._attr_embedding_dim = attr_embedding_dim

        # Create embedders for the different attribute types
        self._attr_embedders = dict()

        if categorical_attributes is not None:
            self._attr_embedders.update(
                construct_categorical_embedders(node_types, attr_embedding_dim, categorical_attributes))

        if continuous_attributes is not None:
            self._attr_embedders.update(
                construct_continuous_embedders(node_types, attr_embedding_dim, continuous_attributes))

        self._attr_embedders.update(
            construct_non_attribute_embedders(node_types, attr_embedding_dim, categorical_attributes,
                                              continuous_attributes))

    def _build(self, features):
        return tf.concat([embed_type(features, len(self._node_types), self._type_embedding_dim),
                          embed_attribute(features, self._attr_embedders, self._attr_embedding_dim)], axis=1)


class RoleEmbedder(snt.AbstractModule):
    def __init__(self, num_edge_types, type_embedding_dim, name="RoleEmbedder"):
        super(RoleEmbedder, self).__init__(name=name)
        self._num_edge_types = num_edge_types
        self._type_embedding_dim = type_embedding_dim

    def _build(self, features):
        return embed_type(features, self._num_edge_types, self._type_embedding_dim)


def embed_type(features, num_types, type_embedding_dim):
    preexistance_feat = tf.expand_dims(tf.cast(features[:, 0], dtype=tf.float32), axis=1)
    type_embedder = snt.Embed(num_types, type_embedding_dim)
    norm = snt.LayerNorm()
    type_embedding = norm(type_embedder(tf.cast(features[:, 1], tf.int32)))
    tf.summary.histogram('type_embedding_histogram', type_embedding)
    return tf.concat([preexistance_feat, type_embedding], axis=1)


def embed_attribute(features, attr_encoders, attr_embedding_dim):
    typewise_attribute_encoder = TypewiseEncoder(attr_encoders, attr_embedding_dim)
    attr_embedding = typewise_attribute_encoder(features[:, 1:])
    tf.summary.histogram('attribute_embedding_histogram', attr_embedding)
    return attr_embedding


def construct_categorical_embedders(node_types, attr_embedding_dim, categorical_attributes):
    attr_embedders = dict()

    # Construct attribute embedders
    for attribute_type, categories in categorical_attributes.items():

        attr_typ_index = node_types.index(attribute_type)

        def make_embedder():
            return CategoricalAttribute(len(categories), attr_embedding_dim,
                                        name=attribute_type + '_cat_embedder')

        # Record the embedder, and the index of the type that it should encode
        attr_embedders[make_embedder] = [attr_typ_index]

    return attr_embedders


def construct_continuous_embedders(node_types, attr_embedding_dim, continuous_attributes):
    attr_embedders = dict()

    # Construct attribute embedders
    for attribute_type in continuous_attributes.keys():

        attr_typ_index = node_types.index(attribute_type)

        def make_embedder():
            return ContinuousAttribute(attr_embedding_dim, name=attribute_type + '_cat_embedder')

        # Record the embedder, and the index of the type that it should encode
        attr_embedders[make_embedder] = [attr_typ_index]

    return attr_embedders


def construct_non_attribute_embedders(node_types, attr_embedding_dim, categorical_attributes, continuous_attributes):

    attribute_names = list(categorical_attributes.keys())
    attribute_names.extend(list(continuous_attributes.keys()))

    non_attribute_nodes = []
    for i, type in enumerate(node_types):
        if type not in attribute_names:
            non_attribute_nodes.append(i)

    # All entities and relations (non-attributes) also need an embedder with matching output dimension, which does
    # nothing. This is provided as a list of their indices
    def make_blank_embedder():
        return BlankAttribute(attr_embedding_dim)

    attr_embedders = dict()

    if len(non_attribute_nodes) > 0:
        attr_embedders[make_blank_embedder] = non_attribute_nodes
    return attr_embedders



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: session/Answer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from grakn_protocol.session import Concept_pb2 as session_dot_Concept__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='session/Answer.proto',
  package='session',
  syntax='proto3',
  serialized_options=_b('\n\026grakn.protocol.sessionB\013AnswerProto'),
  serialized_pb=_b('\n\x14session/Answer.proto\x12\x07session\x1a\x15session/Concept.proto\"\x8d\x03\n\x06\x41nswer\x12\x38\n\x0b\x61nswerGroup\x18\x01 \x01(\x0b\x32\x14.session.AnswerGroupH\x00R\x0b\x61nswerGroup\x12\x35\n\nconceptMap\x18\x02 \x01(\x0b\x32\x13.session.ConceptMapH\x00R\nconceptMap\x12\x38\n\x0b\x63onceptList\x18\x03 \x01(\x0b\x32\x14.session.ConceptListH\x00R\x0b\x63onceptList\x12\x35\n\nconceptSet\x18\x04 \x01(\x0b\x32\x13.session.ConceptSetH\x00R\nconceptSet\x12J\n\x11\x63onceptSetMeasure\x18\x05 \x01(\x0b\x32\x1a.session.ConceptSetMeasureH\x00R\x11\x63onceptSetMeasure\x12&\n\x05value\x18\x06 \x01(\x0b\x32\x0e.session.ValueH\x00R\x05value\x12#\n\x04void\x18\x07 \x01(\x0b\x32\r.session.VoidH\x00R\x04voidB\x08\n\x06\x61nswer\"\x89\x01\n\x0b\x45xplanation\x1a<\n\x03Req\x12\x35\n\x0b\x65xplainable\x18\x01 \x01(\x0b\x32\x13.session.ConceptMapR\x0b\x65xplainable\x1a<\n\x03Res\x12\x35\n\x0b\x65xplanation\x18\x01 \x03(\x0b\x32\x13.session.ConceptMapR\x0b\x65xplanation\"`\n\x0b\x41nswerGroup\x12&\n\x05owner\x18\x01 \x01(\x0b\x32\x10.session.ConceptR\x05owner\x12)\n\x07\x61nswers\x18\x02 \x03(\x0b\x32\x0f.session.AnswerR\x07\x61nswers\"\xc8\x01\n\nConceptMap\x12.\n\x03map\x18\x01 \x03(\x0b\x32\x1c.session.ConceptMap.MapEntryR\x03map\x12\x18\n\x07pattern\x18\x02 \x01(\tR\x07pattern\x12&\n\x0ehasExplanation\x18\x03 \x01(\x08R\x0ehasExplanation\x1aH\n\x08MapEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x10.session.ConceptR\x05value:\x02\x38\x01\"6\n\x0b\x43onceptList\x12\'\n\x04list\x18\x01 \x01(\x0b\x32\x13.session.ConceptIdsR\x04list\"3\n\nConceptSet\x12%\n\x03set\x18\x01 \x01(\x0b\x32\x13.session.ConceptIdsR\x03set\"m\n\x11\x43onceptSetMeasure\x12%\n\x03set\x18\x01 \x01(\x0b\x32\x13.session.ConceptIdsR\x03set\x12\x31\n\x0bmeasurement\x18\x02 \x01(\x0b\x32\x0f.session.NumberR\x0bmeasurement\"0\n\x05Value\x12\'\n\x06number\x18\x01 \x01(\x0b\x32\x0f.session.NumberR\x06number\" \n\x04Void\x12\x18\n\x07message\x18\x01 \x01(\tR\x07message\"\x1e\n\nConceptIds\x12\x10\n\x03ids\x18\x01 \x03(\tR\x03ids\"\x1e\n\x06Number\x12\x14\n\x05value\x18\x01 \x01(\tR\x05valueB%\n\x16grakn.protocol.sessionB\x0b\x41nswerProtob\x06proto3')
  ,
  dependencies=[session_dot_Concept__pb2.DESCRIPTOR,])




_ANSWER = _descriptor.Descriptor(
  name='Answer',
  full_name='session.Answer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='answerGroup', full_name='session.Answer.answerGroup', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='answerGroup', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conceptMap', full_name='session.Answer.conceptMap', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='conceptMap', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conceptList', full_name='session.Answer.conceptList', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='conceptList', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conceptSet', full_name='session.Answer.conceptSet', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='conceptSet', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conceptSetMeasure', full_name='session.Answer.conceptSetMeasure', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='conceptSetMeasure', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='session.Answer.value', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='void', full_name='session.Answer.void', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='void', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='answer', full_name='session.Answer.answer',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=57,
  serialized_end=454,
)


_EXPLANATION_REQ = _descriptor.Descriptor(
  name='Req',
  full_name='session.Explanation.Req',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='explainable', full_name='session.Explanation.Req.explainable', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='explainable', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=472,
  serialized_end=532,
)

_EXPLANATION_RES = _descriptor.Descriptor(
  name='Res',
  full_name='session.Explanation.Res',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='explanation', full_name='session.Explanation.Res.explanation', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='explanation', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=534,
  serialized_end=594,
)

_EXPLANATION = _descriptor.Descriptor(
  name='Explanation',
  full_name='session.Explanation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_EXPLANATION_REQ, _EXPLANATION_RES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=457,
  serialized_end=594,
)


_ANSWERGROUP = _descriptor.Descriptor(
  name='AnswerGroup',
  full_name='session.AnswerGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='owner', full_name='session.AnswerGroup.owner', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='owner', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='answers', full_name='session.AnswerGroup.answers', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='answers', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=596,
  serialized_end=692,
)


_CONCEPTMAP_MAPENTRY = _descriptor.Descriptor(
  name='MapEntry',
  full_name='session.ConceptMap.MapEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='session.ConceptMap.MapEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='session.ConceptMap.MapEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=823,
  serialized_end=895,
)

_CONCEPTMAP = _descriptor.Descriptor(
  name='ConceptMap',
  full_name='session.ConceptMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='map', full_name='session.ConceptMap.map', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='map', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pattern', full_name='session.ConceptMap.pattern', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='pattern', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hasExplanation', full_name='session.ConceptMap.hasExplanation', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='hasExplanation', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CONCEPTMAP_MAPENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=695,
  serialized_end=895,
)


_CONCEPTLIST = _descriptor.Descriptor(
  name='ConceptList',
  full_name='session.ConceptList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='list', full_name='session.ConceptList.list', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='list', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=897,
  serialized_end=951,
)


_CONCEPTSET = _descriptor.Descriptor(
  name='ConceptSet',
  full_name='session.ConceptSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='set', full_name='session.ConceptSet.set', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='set', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=953,
  serialized_end=1004,
)


_CONCEPTSETMEASURE = _descriptor.Descriptor(
  name='ConceptSetMeasure',
  full_name='session.ConceptSetMeasure',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='set', full_name='session.ConceptSetMeasure.set', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='set', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='measurement', full_name='session.ConceptSetMeasure.measurement', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='measurement', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1006,
  serialized_end=1115,
)


_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='session.Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='number', full_name='session.Value.number', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='number', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1117,
  serialized_end=1165,
)


_VOID = _descriptor.Descriptor(
  name='Void',
  full_name='session.Void',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='session.Void.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='message', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1167,
  serialized_end=1199,
)


_CONCEPTIDS = _descriptor.Descriptor(
  name='ConceptIds',
  full_name='session.ConceptIds',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ids', full_name='session.ConceptIds.ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='ids', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1201,
  serialized_end=1231,
)


_NUMBER = _descriptor.Descriptor(
  name='Number',
  full_name='session.Number',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='session.Number.value', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1233,
  serialized_end=1263,
)

_ANSWER.fields_by_name['answerGroup'].message_type = _ANSWERGROUP
_ANSWER.fields_by_name['conceptMap'].message_type = _CONCEPTMAP
_ANSWER.fields_by_name['conceptList'].message_type = _CONCEPTLIST
_ANSWER.fields_by_name['conceptSet'].message_type = _CONCEPTSET
_ANSWER.fields_by_name['conceptSetMeasure'].message_type = _CONCEPTSETMEASURE
_ANSWER.fields_by_name['value'].message_type = _VALUE
_ANSWER.fields_by_name['void'].message_type = _VOID
_ANSWER.oneofs_by_name['answer'].fields.append(
  _ANSWER.fields_by_name['answerGroup'])
_ANSWER.fields_by_name['answerGroup'].containing_oneof = _ANSWER.oneofs_by_name['answer']
_ANSWER.oneofs_by_name['answer'].fields.append(
  _ANSWER.fields_by_name['conceptMap'])
_ANSWER.fields_by_name['conceptMap'].containing_oneof = _ANSWER.oneofs_by_name['answer']
_ANSWER.oneofs_by_name['answer'].fields.append(
  _ANSWER.fields_by_name['conceptList'])
_ANSWER.fields_by_name['conceptList'].containing_oneof = _ANSWER.oneofs_by_name['answer']
_ANSWER.oneofs_by_name['answer'].fields.append(
  _ANSWER.fields_by_name['conceptSet'])
_ANSWER.fields_by_name['conceptSet'].containing_oneof = _ANSWER.oneofs_by_name['answer']
_ANSWER.oneofs_by_name['answer'].fields.append(
  _ANSWER.fields_by_name['conceptSetMeasure'])
_ANSWER.fields_by_name['conceptSetMeasure'].containing_oneof = _ANSWER.oneofs_by_name['answer']
_ANSWER.oneofs_by_name['answer'].fields.append(
  _ANSWER.fields_by_name['value'])
_ANSWER.fields_by_name['value'].containing_oneof = _ANSWER.oneofs_by_name['answer']
_ANSWER.oneofs_by_name['answer'].fields.append(
  _ANSWER.fields_by_name['void'])
_ANSWER.fields_by_name['void'].containing_oneof = _ANSWER.oneofs_by_name['answer']
_EXPLANATION_REQ.fields_by_name['explainable'].message_type = _CONCEPTMAP
_EXPLANATION_REQ.containing_type = _EXPLANATION
_EXPLANATION_RES.fields_by_name['explanation'].message_type = _CONCEPTMAP
_EXPLANATION_RES.containing_type = _EXPLANATION
_ANSWERGROUP.fields_by_name['owner'].message_type = session_dot_Concept__pb2._CONCEPT
_ANSWERGROUP.fields_by_name['answers'].message_type = _ANSWER
_CONCEPTMAP_MAPENTRY.fields_by_name['value'].message_type = session_dot_Concept__pb2._CONCEPT
_CONCEPTMAP_MAPENTRY.containing_type = _CONCEPTMAP
_CONCEPTMAP.fields_by_name['map'].message_type = _CONCEPTMAP_MAPENTRY
_CONCEPTLIST.fields_by_name['list'].message_type = _CONCEPTIDS
_CONCEPTSET.fields_by_name['set'].message_type = _CONCEPTIDS
_CONCEPTSETMEASURE.fields_by_name['set'].message_type = _CONCEPTIDS
_CONCEPTSETMEASURE.fields_by_name['measurement'].message_type = _NUMBER
_VALUE.fields_by_name['number'].message_type = _NUMBER
DESCRIPTOR.message_types_by_name['Answer'] = _ANSWER
DESCRIPTOR.message_types_by_name['Explanation'] = _EXPLANATION
DESCRIPTOR.message_types_by_name['AnswerGroup'] = _ANSWERGROUP
DESCRIPTOR.message_types_by_name['ConceptMap'] = _CONCEPTMAP
DESCRIPTOR.message_types_by_name['ConceptList'] = _CONCEPTLIST
DESCRIPTOR.message_types_by_name['ConceptSet'] = _CONCEPTSET
DESCRIPTOR.message_types_by_name['ConceptSetMeasure'] = _CONCEPTSETMEASURE
DESCRIPTOR.message_types_by_name['Value'] = _VALUE
DESCRIPTOR.message_types_by_name['Void'] = _VOID
DESCRIPTOR.message_types_by_name['ConceptIds'] = _CONCEPTIDS
DESCRIPTOR.message_types_by_name['Number'] = _NUMBER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Answer = _reflection.GeneratedProtocolMessageType('Answer', (_message.Message,), dict(
  DESCRIPTOR = _ANSWER,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.Answer)
  ))
_sym_db.RegisterMessage(Answer)

Explanation = _reflection.GeneratedProtocolMessageType('Explanation', (_message.Message,), dict(

  Req = _reflection.GeneratedProtocolMessageType('Req', (_message.Message,), dict(
    DESCRIPTOR = _EXPLANATION_REQ,
    __module__ = 'session.Answer_pb2'
    # @@protoc_insertion_point(class_scope:session.Explanation.Req)
    ))
  ,

  Res = _reflection.GeneratedProtocolMessageType('Res', (_message.Message,), dict(
    DESCRIPTOR = _EXPLANATION_RES,
    __module__ = 'session.Answer_pb2'
    # @@protoc_insertion_point(class_scope:session.Explanation.Res)
    ))
  ,
  DESCRIPTOR = _EXPLANATION,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.Explanation)
  ))
_sym_db.RegisterMessage(Explanation)
_sym_db.RegisterMessage(Explanation.Req)
_sym_db.RegisterMessage(Explanation.Res)

AnswerGroup = _reflection.GeneratedProtocolMessageType('AnswerGroup', (_message.Message,), dict(
  DESCRIPTOR = _ANSWERGROUP,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.AnswerGroup)
  ))
_sym_db.RegisterMessage(AnswerGroup)

ConceptMap = _reflection.GeneratedProtocolMessageType('ConceptMap', (_message.Message,), dict(

  MapEntry = _reflection.GeneratedProtocolMessageType('MapEntry', (_message.Message,), dict(
    DESCRIPTOR = _CONCEPTMAP_MAPENTRY,
    __module__ = 'session.Answer_pb2'
    # @@protoc_insertion_point(class_scope:session.ConceptMap.MapEntry)
    ))
  ,
  DESCRIPTOR = _CONCEPTMAP,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.ConceptMap)
  ))
_sym_db.RegisterMessage(ConceptMap)
_sym_db.RegisterMessage(ConceptMap.MapEntry)

ConceptList = _reflection.GeneratedProtocolMessageType('ConceptList', (_message.Message,), dict(
  DESCRIPTOR = _CONCEPTLIST,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.ConceptList)
  ))
_sym_db.RegisterMessage(ConceptList)

ConceptSet = _reflection.GeneratedProtocolMessageType('ConceptSet', (_message.Message,), dict(
  DESCRIPTOR = _CONCEPTSET,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.ConceptSet)
  ))
_sym_db.RegisterMessage(ConceptSet)

ConceptSetMeasure = _reflection.GeneratedProtocolMessageType('ConceptSetMeasure', (_message.Message,), dict(
  DESCRIPTOR = _CONCEPTSETMEASURE,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.ConceptSetMeasure)
  ))
_sym_db.RegisterMessage(ConceptSetMeasure)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), dict(
  DESCRIPTOR = _VALUE,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.Value)
  ))
_sym_db.RegisterMessage(Value)

Void = _reflection.GeneratedProtocolMessageType('Void', (_message.Message,), dict(
  DESCRIPTOR = _VOID,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.Void)
  ))
_sym_db.RegisterMessage(Void)

ConceptIds = _reflection.GeneratedProtocolMessageType('ConceptIds', (_message.Message,), dict(
  DESCRIPTOR = _CONCEPTIDS,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.ConceptIds)
  ))
_sym_db.RegisterMessage(ConceptIds)

Number = _reflection.GeneratedProtocolMessageType('Number', (_message.Message,), dict(
  DESCRIPTOR = _NUMBER,
  __module__ = 'session.Answer_pb2'
  # @@protoc_insertion_point(class_scope:session.Number)
  ))
_sym_db.RegisterMessage(Number)


DESCRIPTOR._options = None
_CONCEPTMAP_MAPENTRY._options = None
# @@protoc_insertion_point(module_scope)
